/*=========================================================================

  Program:   Visualization Toolkit
  Module:    vtkVolumeRayCastMIPFunction.h

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
// .NAME vtkVolumeRayCastMIPFunction - A maximum intensity projection ray caster for volumes
//
// .SECTION Description
// vtkVolumeRayCastMIPFunction is a volume ray cast function that
// computes the maximum value encountered along the ray. This is
// either the maximum scalar value, or the maximum opacity, as
// defined by the MaximizeMethod. The color and opacity returned
// by this function is based on the color, scalar opacity, and
// gradient opacity transfer functions defined in the vtkVolumeProperty
// of the vtkVolume.
//
// .SECTION See Also
// vtkVolumeRayCastFunction vtkVolumeRayCastMapper vtkVolumeProperty
// vtkVolumeRayCastCompositeFunction vtkVolumeRayCastIsosurfaceFunction
// vtkVolume vtkVolumeProperty

#ifndef __vtkVolumeRayCastMinIPFunction_h
#define __vtkVolumeRayCastMinIPFunction_h
#include "vtkVolumeRayCastFunction.h"

#define VTK_MINIMIZE_SCALAR_VALUE 0
#define VTK_MINIMIZE_OPACITY      1

class vtkVolumeRayCastMinIPFunction : public vtkVolumeRayCastFunction
{
public:
  static vtkVolumeRayCastMinIPFunction *New();
  vtkTypeMacro(vtkVolumeRayCastMinIPFunction,vtkVolumeRayCastFunction);
  void PrintSelf( ostream& os, vtkIndent indent );


  // Description:
  // Get the scalar value below which all scalar values have zero opacity.
  float GetZeroOpacityThreshold( vtkVolume *vol );


  // Description:
  // Set the MaximizeMethod to either ScalarValue or Opacity.
  vtkSetClampMacro( MinimizeMethod, int,
        VTK_MINIMIZE_SCALAR_VALUE, VTK_MINIMIZE_OPACITY );
  vtkGetMacro(MinimizeMethod,int);
  void SetMinimizeMethodToScalarValue() 
    {this->SetMinimizeMethod(VTK_MINIMIZE_SCALAR_VALUE);}
  void SetMinimizeMethodToOpacity() 
    {this->SetMinimizeMethod(VTK_MINIMIZE_OPACITY);}
  const char *GetMinimizeMethodAsString(void);

//BTX
  void CastRay( vtkVolumeRayCastDynamicInfo *dynamicInfo,
                vtkVolumeRayCastStaticInfo *staticInfo );
//ETX


protected:
  vtkVolumeRayCastMinIPFunction();
  ~vtkVolumeRayCastMinIPFunction();

  int MinimizeMethod;

//BTX
  void SpecificFunctionInitialize( vtkRenderer *ren,
                                   vtkVolume   *vol,
                                   vtkVolumeRayCastStaticInfo *staticInfo,
                                   vtkVolumeRayCastMapper *mapper );

//ETX
private:
  vtkVolumeRayCastMinIPFunction(const vtkVolumeRayCastMinIPFunction&);  // Not implemented.
  void operator=(const vtkVolumeRayCastMinIPFunction&);  // Not implemented.
};



#endif
